#!/usr/bin/python

__pychecker__ = 'unusednames=system,nodes'

import netsim


class Sender(object):
    def __init__(self, destinations):
        self.destinations = destinations
        self.log = None

    def start(self, system, id):
        for i, destination in enumerate(self.destinations):
            system.sendMessage(destination, "message %d -> %d (sequence %d)" % (id, destination, i))


class Receiver(object):
    def __init__(self):
        self.log = []

    def start(self, system, id):
        print "Receiver starting with id", id
        self.id = id

    def messageReceived(self, system, source, message):
        print "Receiver", self.id, "got message from", source, "=", message
        self.log.append(message)


class Forwarder(object):
    def __init__(self, destinations):
        self.destinations = destinations

    def start(self, system, id):
        self.id = id

    def messageReceived(self, system, source, message):
        for destination in self.destinations:
            print "Forwarder", self.id, "got message from", source, " forwarding to", destination
            system.sendMessage(destination, message)


class Validator(object):
    def __init__(self):
        pass

    def validate(self, nodes):
        print "validating state ..."
        for node_id, node in enumerate(nodes):
            print node_id, "=", node.log


def oneSenderTwoReceivers(system):
    """One partial order."""

    r1 = system.addNode(Receiver())
    r2 = system.addNode(Receiver())
    system.addNode(Sender([r1, r2]))

def twoSendersOneReceiver(system):
    """Two partial orders."""

    r = system.addNode(Receiver())
    system.addNode(Sender([r]))
    system.addNode(Sender([r]))

def threeForward(system):
    """Two partial orders."""

    r = system.addNode(Receiver())
    f = system.addNode(Forwarder([r]))
    system.addNode(Sender([f, r]))

def forwardDiamond(system):
    """Two partial orders."""
    
    r = system.addNode(Receiver())
    f1 = system.addNode(Forwarder([r]))
    f2 = system.addNode(Forwarder([r]))
    system.addNode(Sender([f1, f2]))


class Counter(object):
    def __init__(self, destination, count):
        self.destination = destination
        self.desired_count = count
        self.count = 0

    def start(self, system, id):
        self.id = id

    def messageReceived(self, system, source, message):
        self.count += 1
        if self.count != self.desired_count:
            print "Counter", self.id, "got message from", source, "count =", self.count, message
        else:
            print "Counter", self.id, "got message from", source, "forwarding to", self.destination, message
            system.sendMessage(self.destination, "fwd " + message)


def statefulNetwork(system):
    """4 partial orders that depend on state."""

    r = system.addNode(Receiver())
    c = system.addNode(Counter(r, 2))
    system.addNode(Sender([r, c, c]))


def exploderNetwork(system):
    """6 partial orders"""
    r = system.addNode(Receiver())
    f = system.addNode(Forwarder([r, r]))
    system.addNode(Sender([r, f]))


def main2():
    def validate(nodes):
        print "validate"
        return True

    #~ netsim.simulateAll(oneSenderTwoReceivers, validate)
    #~ netsim.simulateAll(twoSendersOneReceiver, validate)
    #~ netsim.simulateAll(forwardDiamond, validate)
    #~ netsim.simulateAll(threeForward, validate)
    #~ netsim.simulateAll(statefulNetwork, validate)
    netsim.simulateAll(exploderNetwork, validate)


def createNetwork(system):
    r = Receiver()
    receiver_id = system.addNode(r)
    s1 = Sender([receiver_id])
    system.addNode(s1)
    s2 = Sender([receiver_id])
    system.addNode(s2)


def main():
    system = netsim.System()
    createNetwork(system)
    system.simulateDeterministic()

    netsim.simulateAll(createNetwork, Validator().validate)


if __name__ == "__main__":
    main()
